// reorder isotable simplex vertices for consistent orientation

#include <fstream>
#include <string>
#include <vector>

#include <CGAL/MP_Float.h>
#include <CGAL/Lazy_exact_nt.h>
#include <CGAL/Cartesian_d.h>
#include <CGAL/Convex_hull_d.h>
#include <CGAL/enum.h>
#include <CGAL/predicates_d.h>

#include "ijktable.h"

using namespace std;

typedef CGAL::Cartesian_d < CGAL::Lazy_exact_nt < CGAL::Quotient <CGAL::MP_Float> > >    R;
typedef CGAL::Point_d<R>             Point;

// global variables
char * fname;

// routines
void backup_file(const char * fname);
void parse_command_line(int argc, char **argv);
void memory_exhaustion();
void usage_error();

int main(int argc, char **argv) {

  std::set_new_handler(memory_exhaustion);
  parse_command_line(argc, argv);

  try {

    IJKTABLE::ISOSURFACE_TABLE isotable(3);
    std::ifstream isotable_file(fname);

    isotable.ReadText(isotable_file);
    if (!isotable.CheckTable()) {
      cerr << "Error found in isosurface table "
	   << fname << "." << endl;
      exit(12);
    };

    isotable_file.close();

    if (isotable.Dimension() < 2) {
      cout << "Isotable dimension is " << isotable.Dimension() << "." << endl;
      cout << "No change to isosurface table." << endl;
      exit(0);
    };

    assert(isotable.NumVerticesPerSimplex() > 1);

    float * coord = new float[isotable.Dimension()];

    for (int it = 0; it < isotable.NumTableEntries(); it++) {
      int nums = isotable.NumSimplices(it);
      if (nums <= 0) continue;

      int iv = 0;
      while (iv < isotable.Polyhedron().NumVertices() &&
	     !isotable.VertexSign(it, iv)) 
	{ iv++; };

      if (iv >= isotable.Polyhedron().NumVertices())
	throw IJKTABLE::ISOSURFACE_TABLE_ERROR("Missing positive vertex.");

      for (int ic = 0; ic < isotable.Dimension(); ic++)
	coord[ic] = isotable.Polyhedron().VertexCoord(iv, ic);

      Point q0(isotable.Dimension(), coord, coord+isotable.Dimension());

      for (int is = 0; is < nums; is++) {

	vector<Point> p;

	// add polyhedron vertex to vector p
	p.push_back(q0);

	for (int jv = 0; jv < isotable.NumVerticesPerSimplex(); jv++) {
	  int ie = isotable.SimplexVertex(it, is, jv);
	  for (int ic = 0; ic < isotable.Dimension(); ic++)
	    coord[ic] = isotable.Polyhedron().MidpointCoord(ie, ic);

	  Point q(isotable.Dimension(), coord, coord+isotable.Dimension());
	  p.push_back(q);
	};

	CGAL::Orientation orient = CGAL::orientation(p.begin(), p.end());

	if (orient == CGAL::NEGATIVE) {
	  // swap last two vertices
	  int j1 = isotable.NumVerticesPerSimplex()-1;
	  int j0 = j1 - 1;
	  int je0 = isotable.SimplexVertex(it, is, j0);
	  int je1 = isotable.SimplexVertex(it, is, j1);
	  isotable.SetSimplexVertex(it, is, j0, je1);
	  isotable.SetSimplexVertex(it, is, j1, je0);
	}
	else if (orient == CGAL::ZERO) {
	  cerr << "Error in table entry " << it << ", simplex " 
	       << is << "." << endl;
	  cerr << "  Polyhedron vertex " << iv 
	       << " is contained in hyperplane of simplex " << is 
	       << "." << endl;
	  cerr << "  Unable to determine simplex orientation." << endl;
	  cerr << "  Continuing." << endl;
	};
      };
    };

    backup_file(fname);

    std::ofstream isotable_file_out(fname);
    isotable.WriteText(isotable_file_out);

    delete [] coord;
  }
  catch (IJKTABLE::ISOSURFACE_TABLE_ERROR error) {
    cerr << "Programming error." << endl;
    cerr << error.Msg() << endl;
    cerr << "Exiting." << endl;
    exit(20);
  };
}

void backup_file(const char * fname)
{
  std::string command;
  command = "mv " + std::string(fname) + " " + std::string(fname) + "~";
  system(command.c_str());
}

void parse_command_line(int argc, char **argv)
{
  if (argc != 2) { usage_error(); };

  fname = argv[1];
}

void memory_exhaustion()
{
   cerr << "Error: Out of memory.  Terminating program." << endl;
   exit(10);
}

void usage_error()
{
  cerr << "Usage: orientisotable {isotable file}" << endl;
  exit(10);
}
